@@ -461,6 +461,9 @@ I32 p5_get_type(PerlInterpreter *my_perl, SV *sv) {
     else if (p5_is_scalar_ref(my_perl, sv)) {
         return 10;
     }
+    else if (SvTYPE(sv) == SVt_PVGV) {
+        return 11;
+    }
     else {
         return 0;
     }
@@ -597,13 +600,53 @@ GV *p5_look_up_package_method(PerlInterpreter *my_perl, char *module, char *name
     PERL_SET_CONTEXT(my_perl);
     {
         HV * const pkg = gv_stashpvn(module, strlen(module), 0);
-        GV * const gv = Perl_gv_fetchmethod_autoload(aTHX_ pkg, name, TRUE);
+        GV * const gv = gv_fetchmeth_pvn(pkg, name, strlen(name), -1, SVf_UTF8);
         if (gv && isGV(gv))
             return gv;
         return NULL;
     }
 }
 
+SV *p5_call_inherited_package_method(PerlInterpreter *my_perl, char *package, char *base_package, char *name, int len, SV *args[], I32 *count, I32 *err, I32 *type) {
+    PERL_SET_CONTEXT(my_perl);
+    {
+        dSP;
+        SV * retval = NULL;
+        HV * stash = gv_stashpvn(package, strlen(package), SVf_UTF8);
+        int flags = G_ARRAY | G_EVAL;
+
+        if (stash == NULL) {
+            *type = -1; /* signal that a wrapper package needs to be created */
+            return NULL;
+        }
+
+        ENTER;
+        SAVETMPS;
+
+        PUSHMARK(SP);
+
+        XPUSHs(newSVpv(package, 0));
+        push_arguments(sp, len, args);
+
+        {
+            GV * const gv = p5_look_up_package_method(my_perl, base_package, name);
+            SV * const rv = sv_2mortal(newRV((SV*)GvCV(gv)));
+            *count = call_sv(rv, flags);
+        }
+
+        SPAGAIN;
+
+        handle_p5_error(err);
+
+        retval = pop_return_values(my_perl, sp, *count, type);
+
+        FREETMPS;
+        LEAVE;
+
+        return retval;
+    }
+}
+
 GV *p5_look_up_method(PerlInterpreter *my_perl, SV *obj, char *name) {
     PERL_SET_CONTEXT(my_perl);
     {
@@ -620,13 +663,12 @@ char *p5_stash_name(PerlInterpreter *my_perl, SV *obj) {
     return HvNAME(pkg);
 }
 
-SV *p5_call_gv(PerlInterpreter *my_perl, SV *obj, I32 context, GV *gv, int len, SV *args[], I32 *count, I32 *err, I32 *type) {
+SV *p5_call_gv(PerlInterpreter *my_perl, GV *gv, int len, SV *args[], I32 *count, I32 *err, I32 *type) {
     PERL_SET_CONTEXT(my_perl);
     {
         dSP;
         int i;
         SV * retval = NULL;
-        int flags = (context ? G_SCALAR : G_ARRAY) | G_EVAL;
 
         ENTER;
         SAVETMPS;
@@ -646,9 +688,9 @@ SV *p5_call_gv(PerlInterpreter *my_perl, SV *obj, I32 context, GV *gv, int len,
 
         PUTBACK;
 
-        SV * const rv = sv_2mortal(newRV((SV*)GvCV(gv)));
+        SV * const rv = sv_2mortal(newRV((SV*)GvCV(gv))); /* FIXME: can be done once */
 
-        *count = call_sv(rv, flags);
+        *count = call_sv(rv, G_ARRAY | G_EVAL);
         SPAGAIN;
 
         handle_p5_error(err);
@@ -663,6 +705,118 @@ SV *p5_call_gv(PerlInterpreter *my_perl, SV *obj, I32 context, GV *gv, int len,
     }
 }
 
+SV *p5_scalar_call_gv(PerlInterpreter *my_perl, GV *gv, int len, SV *args[], I32 *count, I32 *err, I32 *type) {
+    PERL_SET_CONTEXT(my_perl);
+    {
+        dSP;
+        int i;
+        SV * retval = NULL;
+
+        ENTER;
+        SAVETMPS;
+
+        PUSHMARK(SP);
+
+        if (len > 1) {
+            XPUSHs(args[0]);
+            for (i = 1; i < len; i++) {
+                if (args[i] != NULL) /* skip Nil which gets turned into NULL */
+                    XPUSHs(sv_2mortal(args[i]));
+            }
+        }
+        else if (len > 0)
+            if (args != NULL) /* skip Nil which gets turned into NULL */
+                XPUSHs((SV*)args);
+
+        PUTBACK;
+
+        SV * const rv = sv_2mortal(newRV((SV*)GvCV(gv))); /* FIXME: can be done once */
+
+        *count = call_sv(rv, G_SCALAR | G_EVAL);
+        SPAGAIN;
+
+        handle_p5_error(err);
+        retval = pop_return_values(my_perl, sp, *count, type);
+        SPAGAIN;
+
+        PUTBACK;
+        FREETMPS;
+        LEAVE;
+
+        return retval;
+    }
+}
+
+SV *p5_call_gv_two_args(PerlInterpreter *my_perl, GV *gv, SV *arg, SV *arg2, I32 *count, I32 *type, I32 *err) {
+    PERL_SET_CONTEXT(my_perl);
+    {
+        dSP;
+        SV * retval = NULL;
+
+        ENTER;
+        SAVETMPS;
+
+        PUSHMARK(SP);
+
+        XPUSHs((SV*)arg);
+        XPUSHs((SV*)arg2);
+
+        PUTBACK;
+
+        SV * const rv = sv_2mortal(newRV((SV*)GvCV(gv)));
+
+        *count = call_sv(rv, G_ARRAY | G_EVAL);
+        SPAGAIN;
+
+        handle_p5_error(err);
+        if (*err)
+            fprintf(stderr, "err: %d\n", *err);
+        retval = pop_return_values(my_perl, sp, *count, type);
+        SPAGAIN;
+
+        PUTBACK;
+        FREETMPS;
+        LEAVE;
+
+        return retval;
+    }
+}
+
+SV *p5_scalar_call_gv_two_args(PerlInterpreter *my_perl, GV *gv, SV *arg, SV *arg2, I32 *count, I32 *type, I32 *err) {
+    PERL_SET_CONTEXT(my_perl);
+    {
+        dSP;
+        SV * retval = NULL;
+
+        ENTER;
+        SAVETMPS;
+
+        PUSHMARK(SP);
+
+        XPUSHs((SV*)arg);
+        XPUSHs((SV*)arg2);
+
+        PUTBACK;
+
+        SV * const rv = sv_2mortal(newRV((SV*)GvCV(gv)));
+
+        *count = call_sv(rv, G_SCALAR | G_EVAL);
+        SPAGAIN;
+
+        handle_p5_error(err);
+        if (*err)
+            fprintf(stderr, "err: %d\n", *err);
+        retval = pop_return_values(my_perl, sp, *count, type);
+        SPAGAIN;
+
+        PUTBACK;
+        FREETMPS;
+        LEAVE;
+
+        return retval;
+    }
+}
+
 SV *p5_call_method(PerlInterpreter *my_perl, SV *obj, I32 context, char *name, int len, SV *args[], I32 *count, I32 *err, I32 *type) {
     PERL_SET_CONTEXT(my_perl);
     {
@@ -887,6 +1041,19 @@ void p5_rebless_object(PerlInterpreter *my_perl, SV *obj, char *package, IV i) {
     }
 }
 
+void p5_add_magic(PerlInterpreter *my_perl, SV *obj, IV i) {
+    PERL_SET_CONTEXT(my_perl);
+    {
+        SV * const inst = SvRV(obj);
+        _perl6_magic priv;
+
+        /* set up magic */
+        priv.key = PERL6_MAGIC_KEY;
+        priv.index = i;
+        sv_magicext(inst, inst, PERL_MAGIC_ext, &p5_inline_mg_vtbl, (char *) &priv, sizeof(priv));
+    }
+}
+
 SV *p5_wrap_p6_object(PerlInterpreter *my_perl, IV i, SV *p5obj) {
     PERL_SET_CONTEXT(my_perl);
     {
@@ -1109,6 +1276,10 @@ XS(p5_call_p6_method) {
     }
     SV * const obj_deref = SvRV(obj);
     MAGIC * const mg = mg_find(obj_deref, '~');
+    if (!mg) {
+        XSRETURN_EMPTY;
+        return;
+    }
     _perl6_magic* const p6mg = (_perl6_magic*)(mg->mg_ptr);
     SV *err = NULL;
     SV * const args_rv = newRV_noinc((SV *) args);
